import enum

from flask import Flask
from flask_cors import CORS
from datetime import datetime, date
from flask.json import JSONEncoder, jsonify
from importlib.metadata import entry_points
from werkzeug.exceptions import HTTPException

from flaschengeist import logger
from flaschengeist.controller import pluginController
from flaschengeist.utils.hook import Hook
from flaschengeist.config import configure_app


class CustomJSONEncoder(JSONEncoder):
    def default(self, o):
        try:
            # Check if custom model
            return o.serialize()
        except AttributeError:
            pass
        if isinstance(o, datetime) or isinstance(o, date):
            return o.isoformat()
        if isinstance(o, enum.Enum):
            return o.value
        try:
            # Check if iterable
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)
        return JSONEncoder.default(self, o)


@Hook("plugins.loaded")
def load_plugins(app: Flask):
    app.config["FG_PLUGINS"] = {}
    all_plugins = entry_points(group="flaschengeist.plugins")

    for plugin in pluginController.get_enabled_plugins():
        logger.debug(f"Searching for enabled plugin {plugin.name}")
        entry_point = all_plugins.select(name=plugin.name)
        if not entry_point:
            logger.error(
                f"Plugin {plugin.name} was enabled, but could not be found.",
                exc_info=True,
            )
            continue
        try:
            loaded = entry_point[0].load()(entry_point[0])
            if hasattr(plugin, "blueprint") and plugin.blueprint is not None:
                app.register_blueprint(plugin.blueprint)
        except:
            logger.error(
                f"Plugin {plugin.name} was enabled, but could not be loaded due to an error.",
                exc_info=True,
            )
            continue
        logger.info(f"Loaded plugin: {plugin.name}")
        app.config["FG_PLUGINS"][plugin.name] = loaded


def create_app(test_config=None, cli=False):
    app = Flask("flaschengeist")
    app.json_encoder = CustomJSONEncoder
    CORS(app)

    with app.app_context():
        from flaschengeist.database import db, migrate

        configure_app(app, test_config)
        db.init_app(app)
        migrate.init_app(app, db, compare_type=True)
        load_plugins(app)

    @app.route("/", methods=["GET"])
    def __get_state():
        from . import __version__ as version

        return jsonify({"plugins": app.config["FG_PLUGINS"], "version": version})

    @app.errorhandler(Exception)
    def handle_exception(e):
        if isinstance(e, HTTPException):
            logger.debug(e.description, exc_info=True)
            return jsonify({"error": e.description}), e.code
        if isinstance(e, OperationalError):
            logger.error(e, exc_info=True)
            return {"error": "Database unavailable"}, 504
        logger.error(str(e), exc_info=True)
        return jsonify({"error": "Internal server error occurred"}), 500

    return app
